@@ -1,6 +1,6 @@
 /***
  * ASM: a very small and fast Java bytecode manipulation framework
- * Copyright (c) 2000-2005 INRIA, France Telecom
+ * Copyright (c) 2000-2007 INRIA, France Telecom
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -81,127 +81,129 @@ import org.objectweb.asm.Type;
  */
 public class GeneratorAdapter extends LocalVariablesSorter {
 
-    private final static Type BYTE_TYPE = Type.getType("Ljava/lang/Byte;");
+    private static final String CLDESC = "Ljava/lang/Class;";
 
-    private final static Type BOOLEAN_TYPE = Type.getType("Ljava/lang/Boolean;");
+    private static final Type BYTE_TYPE = Type.getObjectType("java/lang/Byte");
 
-    private final static Type SHORT_TYPE = Type.getType("Ljava/lang/Short;");
+    private static final Type BOOLEAN_TYPE = Type.getObjectType("java/lang/Boolean");
 
-    private final static Type CHARACTER_TYPE = Type.getType("Ljava/lang/Character;");
+    private static final Type SHORT_TYPE = Type.getObjectType("java/lang/Short");
 
-    private final static Type INTEGER_TYPE = Type.getType("Ljava/lang/Integer;");
+    private static final Type CHARACTER_TYPE = Type.getObjectType("java/lang/Character");
 
-    private final static Type FLOAT_TYPE = Type.getType("Ljava/lang/Float;");
+    private static final Type INTEGER_TYPE = Type.getObjectType("java/lang/Integer");
 
-    private final static Type LONG_TYPE = Type.getType("Ljava/lang/Long;");
+    private static final Type FLOAT_TYPE = Type.getObjectType("java/lang/Float");
 
-    private final static Type DOUBLE_TYPE = Type.getType("Ljava/lang/Double;");
+    private static final Type LONG_TYPE = Type.getObjectType("java/lang/Long");
 
-    private final static Type NUMBER_TYPE = Type.getType("Ljava/lang/Number;");
+    private static final Type DOUBLE_TYPE = Type.getObjectType("java/lang/Double");
 
-    private final static Type OBJECT_TYPE = Type.getType("Ljava/lang/Object;");
+    private static final Type NUMBER_TYPE = Type.getObjectType("java/lang/Number");
 
-    private final static Method BOOLEAN_VALUE = Method.getMethod("boolean booleanValue()");
+    private static final Type OBJECT_TYPE = Type.getObjectType("java/lang/Object");
 
-    private final static Method CHAR_VALUE = Method.getMethod("char charValue()");
+    private static final Method BOOLEAN_VALUE = Method.getMethod("boolean booleanValue()");
 
-    private final static Method INT_VALUE = Method.getMethod("int intValue()");
+    private static final Method CHAR_VALUE = Method.getMethod("char charValue()");
 
-    private final static Method FLOAT_VALUE = Method.getMethod("float floatValue()");
+    private static final Method INT_VALUE = Method.getMethod("int intValue()");
 
-    private final static Method LONG_VALUE = Method.getMethod("long longValue()");
+    private static final Method FLOAT_VALUE = Method.getMethod("float floatValue()");
 
-    private final static Method DOUBLE_VALUE = Method.getMethod("double doubleValue()");
+    private static final Method LONG_VALUE = Method.getMethod("long longValue()");
+
+    private static final Method DOUBLE_VALUE = Method.getMethod("double doubleValue()");
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int ADD = Opcodes.IADD;
+    public static final int ADD = Opcodes.IADD;
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int SUB = Opcodes.ISUB;
+    public static final int SUB = Opcodes.ISUB;
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int MUL = Opcodes.IMUL;
+    public static final int MUL = Opcodes.IMUL;
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int DIV = Opcodes.IDIV;
+    public static final int DIV = Opcodes.IDIV;
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int REM = Opcodes.IREM;
+    public static final int REM = Opcodes.IREM;
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int NEG = Opcodes.INEG;
+    public static final int NEG = Opcodes.INEG;
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int SHL = Opcodes.ISHL;
+    public static final int SHL = Opcodes.ISHL;
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int SHR = Opcodes.ISHR;
+    public static final int SHR = Opcodes.ISHR;
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int USHR = Opcodes.IUSHR;
+    public static final int USHR = Opcodes.IUSHR;
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int AND = Opcodes.IAND;
+    public static final int AND = Opcodes.IAND;
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int OR = Opcodes.IOR;
+    public static final int OR = Opcodes.IOR;
 
     /**
      * Constant for the {@link #math math} method.
      */
-    public final static int XOR = Opcodes.IXOR;
+    public static final int XOR = Opcodes.IXOR;
 
     /**
      * Constant for the {@link #ifCmp ifCmp} method.
      */
-    public final static int EQ = Opcodes.IFEQ;
+    public static final int EQ = Opcodes.IFEQ;
 
     /**
      * Constant for the {@link #ifCmp ifCmp} method.
      */
-    public final static int NE = Opcodes.IFNE;
+    public static final int NE = Opcodes.IFNE;
 
     /**
      * Constant for the {@link #ifCmp ifCmp} method.
      */
-    public final static int LT = Opcodes.IFLT;
+    public static final int LT = Opcodes.IFLT;
 
     /**
      * Constant for the {@link #ifCmp ifCmp} method.
      */
-    public final static int GE = Opcodes.IFGE;
+    public static final int GE = Opcodes.IFGE;
 
     /**
      * Constant for the {@link #ifCmp ifCmp} method.
      */
-    public final static int GT = Opcodes.IFGT;
+    public static final int GT = Opcodes.IFGT;
 
     /**
      * Constant for the {@link #ifCmp ifCmp} method.
      */
-    public final static int LE = Opcodes.IFLE;
+    public static final int LE = Opcodes.IFLE;
 
     /**
      * Access flags of the method visited by this adapter.
@@ -221,7 +223,27 @@ public class GeneratorAdapter extends LocalVariablesSorter {
     /**
      * Types of the local variables of the method visited by this adapter.
      */
-    private final List localTypes;
+    private final List localTypes = new ArrayList();
+
+    /**
+     * Creates a new {@link GeneratorAdapter}.
+     * 
+     * @param mv the method visitor to which this adapter delegates calls.
+     * @param access the method's access flags (see {@link Opcodes}).
+     * @param name the method's name.
+     * @param desc the method's descriptor (see {@link Type Type}).
+     */
+    public GeneratorAdapter(
+        final MethodVisitor mv,
+        final int access,
+        final String name,
+        final String desc)
+    {
+        super(access, desc, mv);
+        this.access = access;
+        this.returnType = Type.getReturnType(desc);
+        this.argumentTypes = Type.getArgumentTypes(desc);
+    }
 
     /**
      * Creates a new {@link GeneratorAdapter}.
@@ -239,7 +261,6 @@ public class GeneratorAdapter extends LocalVariablesSorter {
         this.access = access;
         this.returnType = method.getReturnType();
         this.argumentTypes = method.getArgumentTypes();
-        this.localTypes = new ArrayList();
     }
 
     /**
@@ -377,7 +398,58 @@ public class GeneratorAdapter extends LocalVariablesSorter {
         if (value == null) {
             mv.visitInsn(Opcodes.ACONST_NULL);
         } else {
-            mv.visitLdcInsn(value);
+            switch (value.getSort()) {
+                case Type.BOOLEAN:
+                    mv.visitFieldInsn(Opcodes.GETSTATIC,
+                            "java/lang/Boolean",
+                            "TYPE",
+                            CLDESC);
+                    break;
+                case Type.CHAR:
+                    mv.visitFieldInsn(Opcodes.GETSTATIC,
+                            "java/lang/Char",
+                            "TYPE",
+                            CLDESC);
+                    break;
+                case Type.BYTE:
+                    mv.visitFieldInsn(Opcodes.GETSTATIC,
+                            "java/lang/Byte",
+                            "TYPE",
+                            CLDESC);
+                    break;
+                case Type.SHORT:
+                    mv.visitFieldInsn(Opcodes.GETSTATIC,
+                            "java/lang/Short",
+                            "TYPE",
+                            CLDESC);
+                    break;
+                case Type.INT:
+                    mv.visitFieldInsn(Opcodes.GETSTATIC,
+                            "java/lang/Integer",
+                            "TYPE",
+                            CLDESC);
+                    break;
+                case Type.FLOAT:
+                    mv.visitFieldInsn(Opcodes.GETSTATIC,
+                            "java/lang/Float",
+                            "TYPE",
+                            CLDESC);
+                    break;
+                case Type.LONG:
+                    mv.visitFieldInsn(Opcodes.GETSTATIC,
+                            "java/lang/Long",
+                            "TYPE",
+                            CLDESC);
+                    break;
+                case Type.DOUBLE:
+                    mv.visitFieldInsn(Opcodes.GETSTATIC,
+                            "java/lang/Double",
+                            "TYPE",
+                            CLDESC);
+                    break;
+                default:
+                    mv.visitLdcInsn(value);
+            }
         }
     }
 
@@ -394,7 +466,7 @@ public class GeneratorAdapter extends LocalVariablesSorter {
      *         variables array.
      */
     private int getArgIndex(final int arg) {
-        int index = ((access & Opcodes.ACC_STATIC) == 0 ? 1 : 0);
+        int index = (access & Opcodes.ACC_STATIC) == 0 ? 1 : 0;
         for (int i = 0; i < arg; i++) {
             index += argumentTypes[i].getSize();
         }
@@ -494,48 +566,30 @@ public class GeneratorAdapter extends LocalVariablesSorter {
     // Instructions to load and store local variables
     // ------------------------------------------------------------------------
 
-    /**
-     * Creates a new local variable of the given type.
-     * 
-     * @param type the type of the local variable to be created.
-     * @return the identifier of the newly created local variable.
-     */
-    public int newLocal(final Type type) {
-        int local = super.newLocal(type.getSize());
-        setLocalType(local, type);
-        return local;
-    }
-
     /**
      * Returns the type of the given local variable.
      * 
-     * @param local a local variable identifier, as returned by {@link #newLocal
-     *        newLocal}.
+     * @param local a local variable identifier, as returned by
+     *        {@link LocalVariablesSorter#newLocal(Type) newLocal()}.
      * @return the type of the given local variable.
      */
     public Type getLocalType(final int local) {
         return (Type) localTypes.get(local - firstLocal);
     }
 
-    /**
-     * Sets the current type of the given local variable.
-     * 
-     * @param local a local variable identifier, as returned by {@link #newLocal
-     *        newLocal}.
-     * @param type the type of the value being stored in the local variable
-     */
-    private void setLocalType(final int local, final Type type) {
+    protected void setLocalType(final int local, final Type type) {
         int index = local - firstLocal;
-        while (localTypes.size() < index + 1)
+        while (localTypes.size() < index + 1) {
             localTypes.add(null);
+        }
         localTypes.set(index, type);
     }
 
     /**
      * Generates the instruction to load the given local variable on the stack.
      * 
-     * @param local a local variable identifier, as returned by {@link #newLocal
-     *        newLocal}.
+     * @param local a local variable identifier, as returned by
+     *        {@link LocalVariablesSorter#newLocal(Type) newLocal()}.
      */
     public void loadLocal(final int local) {
         loadInsn(getLocalType(local), local);
@@ -544,8 +598,9 @@ public class GeneratorAdapter extends LocalVariablesSorter {
     /**
      * Generates the instruction to load the given local variable on the stack.
      * 
-     * @param local a local variable identifier, as returned by {@link #newLocal
-     *        newLocal}.
+     * @param local a local variable identifier, as returned by
+     *        {@link LocalVariablesSorter#newLocal(Type) newLocal()}.
+     * @param type the type of this local variable.
      */
     public void loadLocal(final int local, final Type type) {
         setLocalType(local, type);
@@ -556,8 +611,8 @@ public class GeneratorAdapter extends LocalVariablesSorter {
      * Generates the instruction to store the top stack value in the given local
      * variable.
      * 
-     * @param local a local variable identifier, as returned by {@link #newLocal
-     *        newLocal}.
+     * @param local a local variable identifier, as returned by
+     *        {@link LocalVariablesSorter#newLocal(Type) newLocal()}.
      */
     public void storeLocal(final int local) {
         storeInsn(getLocalType(local), local);
@@ -567,8 +622,9 @@ public class GeneratorAdapter extends LocalVariablesSorter {
      * Generates the instruction to store the top stack value in the given local
      * variable.
      * 
-     * @param local a local variable identifier, as returned by {@link #newLocal
-     *        newLocal}.
+     * @param local a local variable identifier, as returned by
+     *        {@link LocalVariablesSorter#newLocal(Type) newLocal()}.
+     * @param type the type of this local variable.
      */
     public void storeLocal(final int local, final Type type) {
         setLocalType(local, type);
@@ -921,15 +977,6 @@ public class GeneratorAdapter extends LocalVariablesSorter {
      */
     public void ifCmp(final Type type, final int mode, final Label label) {
         int intOp = -1;
-        int jumpMode = mode;
-        switch (mode) {
-            case GE:
-                jumpMode = LT;
-                break;
-            case LE:
-                jumpMode = GT;
-                break;
-        }
         switch (type.getSort()) {
             case Type.LONG:
                 mv.visitInsn(Opcodes.LCMP);
@@ -976,6 +1023,15 @@ public class GeneratorAdapter extends LocalVariablesSorter {
                 mv.visitJumpInsn(intOp, label);
                 return;
         }
+        int jumpMode = mode;
+        switch (mode) {
+            case GE:
+                jumpMode = LT;
+                break;
+            case LE:
+                jumpMode = GT;
+                break;
+        }
         mv.visitJumpInsn(jumpMode, label);
     }
 
@@ -1035,8 +1091,8 @@ public class GeneratorAdapter extends LocalVariablesSorter {
     /**
      * Generates a RET instruction.
      * 
-     * @param local a local variable identifier, as returned by {@link #newLocal
-     *        newLocal}.
+     * @param local a local variable identifier, as returned by
+     *        {@link LocalVariablesSorter#newLocal(Type) newLocal()}.
      */
     public void ret(final int local) {
         mv.visitVarInsn(Opcodes.RET, local);
@@ -1274,13 +1330,7 @@ public class GeneratorAdapter extends LocalVariablesSorter {
      * @param type the instruction's operand.
      */
     private void typeInsn(final int opcode, final Type type) {
-        String desc;
-        if (type.getSort() == Type.ARRAY) {
-            desc = type.getDescriptor();
-        } else {
-            desc = type.getInternalName();
-        }
-        mv.visitTypeInsn(opcode, desc);
+        mv.visitTypeInsn(opcode, type.getInternalName());
     }
 
     /**
@@ -1411,6 +1461,7 @@ public class GeneratorAdapter extends LocalVariablesSorter {
         if ((access & Opcodes.ACC_ABSTRACT) == 0) {
             mv.visitMaxs(0, 0);
         }
+        mv.visitEnd();
     }
 
     /**
