@@ -1,6 +1,6 @@
 /***
  * ASM: a very small and fast Java bytecode manipulation framework
- * Copyright (c) 2000-2005 INRIA, France Telecom
+ * Copyright (c) 2000-2007 INRIA, France Telecom
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -29,6 +29,8 @@
  */
 package org.objectweb.asm.tree.analysis;
 
+import java.util.List;
+
 import org.objectweb.asm.Type;
 
 /**
@@ -51,6 +53,11 @@ public class SimpleVerifier extends BasicVerifier {
      */
     private final Type currentSuperClass;
 
+    /**
+     * The interfaces implemented by the class that is verified.
+     */
+    private final List currentClassInterfaces;
+
     /**
      * If the class that is verified is an interface.
      */
@@ -75,20 +82,55 @@ public class SimpleVerifier extends BasicVerifier {
         final Type currentClass,
         final Type currentSuperClass,
         final boolean isInterface)
+    {
+        this(currentClass, currentSuperClass, null, isInterface);
+    }
+
+    /**
+     * Constructs a new {@link SimpleVerifier} to verify a specific class. This
+     * class will not be loaded into the JVM since it may be incorrect.
+     * 
+     * @param currentClass the class that is verified.
+     * @param currentSuperClass the super class of the class that is verified.
+     * @param currentClassInterfaces the interfaces implemented by the class
+     *        that is verified.
+     * @param isInterface if the class that is verified is an interface.
+     */
+    public SimpleVerifier(
+        final Type currentClass,
+        final Type currentSuperClass,
+        final List currentClassInterfaces,
+        final boolean isInterface)
     {
         this.currentClass = currentClass;
         this.currentSuperClass = currentSuperClass;
+        this.currentClassInterfaces = currentClassInterfaces;
         this.isInterface = isInterface;
     }
 
     public Value newValue(final Type type) {
+        if (type == null) {
+            return BasicValue.UNINITIALIZED_VALUE;
+        }
+
+        boolean isArray = type.getSort() == Type.ARRAY;
+        if (isArray) {
+            switch (type.getElementType().getSort()) {
+                case Type.BOOLEAN:
+                case Type.CHAR:
+                case Type.BYTE:
+                case Type.SHORT:
+                    return new BasicValue(type);
+            }
+        }
+
         Value v = super.newValue(type);
         if (v == BasicValue.REFERENCE_VALUE) {
-            if (type.getSort() == Type.ARRAY) {
+            if (isArray) {
                 v = newValue(type.getElementType());
                 String desc = ((BasicValue) v).getType().getDescriptor();
                 for (int i = 0; i < type.getDimensions(); ++i) {
-                    desc = "[" + desc;
+                    desc = '[' + desc;
                 }
                 v = new BasicValue(Type.getType(desc));
             } else {
@@ -100,11 +142,8 @@ public class SimpleVerifier extends BasicVerifier {
 
     protected boolean isArrayValue(final Value value) {
         Type t = ((BasicValue) value).getType();
-        if (t != null) {
-            return t.getDescriptor().equals("Lnull;")
-                    || t.getSort() == Type.ARRAY;
-        }
-        return false;
+        return t != null
+                && ("Lnull;".equals(t.getDescriptor()) || t.getSort() == Type.ARRAY);
     }
 
     protected Value getElementValue(final Value objectArrayValue)
@@ -115,19 +154,16 @@ public class SimpleVerifier extends BasicVerifier {
             if (arrayType.getSort() == Type.ARRAY) {
                 return newValue(Type.getType(arrayType.getDescriptor()
                         .substring(1)));
-            } else if (arrayType.getDescriptor().equals("Lnull;")) {
+            } else if ("Lnull;".equals(arrayType.getDescriptor())) {
                 return objectArrayValue;
             }
         }
-        throw new AnalyzerException("Not an array type");
+        throw new Error("Internal error");
     }
 
     protected boolean isSubTypeOf(final Value value, final Value expected) {
         Type expectedType = ((BasicValue) expected).getType();
         Type type = ((BasicValue) value).getType();
-        if (expectedType == null) {
-            return type == null;
-        }
         switch (expectedType.getSort()) {
             case Type.INT:
             case Type.FLOAT:
@@ -136,11 +172,7 @@ public class SimpleVerifier extends BasicVerifier {
                 return type == expectedType;
             case Type.ARRAY:
             case Type.OBJECT:
-                if (expectedType.getDescriptor().equals("Lnull;")) {
-                    return type.getSort() == Type.OBJECT
-                            || type.getSort() == Type.ARRAY;
-                }
-                if (type.getDescriptor().equals("Lnull;")) {
+                if ("Lnull;".equals(type.getDescriptor())) {
                     return true;
                 } else if (type.getSort() == Type.OBJECT
                         || type.getSort() == Type.ARRAY)
@@ -150,7 +182,7 @@ public class SimpleVerifier extends BasicVerifier {
                     return false;
                 }
             default:
-                throw new RuntimeException("Internal error");
+                throw new Error("Internal error");
         }
     }
 
@@ -164,10 +196,10 @@ public class SimpleVerifier extends BasicVerifier {
                 if (u != null
                         && (u.getSort() == Type.OBJECT || u.getSort() == Type.ARRAY))
                 {
-                    if (t.getDescriptor().equals("Lnull;")) {
+                    if ("Lnull;".equals(t.getDescriptor())) {
                         return w;
                     }
-                    if (u.getDescriptor().equals("Lnull;")) {
+                    if ("Lnull;".equals(u.getDescriptor())) {
                         return v;
                     }
                     if (isAssignableFrom(t, u)) {
@@ -196,14 +228,14 @@ public class SimpleVerifier extends BasicVerifier {
         return v;
     }
 
-    private boolean isInterface(final Type t) {
+    protected boolean isInterface(final Type t) {
         if (currentClass != null && t.equals(currentClass)) {
             return isInterface;
         }
         return getClass(t).isInterface();
     }
 
-    private Type getSuperClass(final Type t) {
+    protected Type getSuperClass(final Type t) {
         if (currentClass != null && t.equals(currentClass)) {
             return currentSuperClass;
         }
@@ -211,15 +243,30 @@ public class SimpleVerifier extends BasicVerifier {
         return c == null ? null : Type.getType(c);
     }
 
-    private boolean isAssignableFrom(final Type t, final Type u) {
+    protected boolean isAssignableFrom(final Type t, final Type u) {
         if (t.equals(u)) {
             return true;
         }
         if (currentClass != null && t.equals(currentClass)) {
-            return isAssignableFrom(t, getSuperClass(u));
+            if (getSuperClass(u) == null) {
+                return false;
+            } else {
+                return isAssignableFrom(t, getSuperClass(u));
+            }
         }
         if (currentClass != null && u.equals(currentClass)) {
-            return isAssignableFrom(t, currentSuperClass);
+            if (isAssignableFrom(t, currentSuperClass)) {
+                return true;
+            }
+            if (currentClassInterfaces != null) {
+                for (int i = 0; i < currentClassInterfaces.size(); ++i) {
+                    Type v = (Type) currentClassInterfaces.get(i);
+                    if (isAssignableFrom(t, v)) {
+                        return true;
+                    }
+                }
+            }
+            return false;
         }
         return getClass(t).isAssignableFrom(getClass(u));
     }
